
import logging, aiohttp, asyncio
from aiohttp import ClientSession
from typing import List, Optional, AsyncGenerator, Final, Sequence, Tuple, Union, Callable, Awaitable, Any, Dict, Literal

from .datatypes import *


async def _run_func(func, *args):
    if hasattr(func, '__call__'):
        func= func.__call__
    if asyncio.iscoroutinefunction(func):
        return await func(*args)
    else:
        return func(*args)

_LOGGER = logging.getLogger("aio-sse-chat")


DEFAULT_VALID_HTTP_CODES: Final[Tuple[int, int, int]] = (200, 301, 307)

EventCallback = Union[Callable[[SSEvent], Any], Callable[[SSEvent], Awaitable[Any]]]
'''Callback when an event is received. It can be async or non-async'''

async def aiosseclient(
    url: str, 
    session: Optional[ClientSession] = None,
    method: Literal['get', 'post', 'put', 'delete', 'patch', 'options', 'head'] = 'get',
    json: Optional[dict] = None,
    headers: Optional[dict[str, str]] = None,
    last_id: Optional[str] = None,
    valid_http_codes: Sequence[int] = DEFAULT_VALID_HTTP_CODES,
    exit_events: Optional[List[str]] = None,
    events: Optional[Dict[str, EventCallback]] = None,
    **kwargs,
) -> AsyncGenerator[SSEvent, None]:
    '''
    A modifier to enable the aio session to handle SSE events.
    
    Args:
        - session: aiohttp.ClientSession
        - url: the url to connect to
        - method: the method to use for the request
        - json: the json data to be sent with the request
        - last_id: the last event id
        - valid_http_codes: the valid http codes, default (200, 301, 307)
        - exit_events: the events that will cause the session to close
        - headers: the extra headers to be added to the request
        - events: Callables for triggering events. {event_name: no-arg-callable}. Callable can be async or non-async.
        - kwargs: other arguments to be passed to the aiohttp.ClientSession.request method
        
    Example:
    ```python
    async for event in aiosseclient(url=some_url, method='post', json=some_data):
        print(data, end='', flush=True)
    ```
    '''
    if session is None:
        session = aiohttp.ClientSession()
    if not isinstance(session, ClientSession):
        raise ValueError('session must be an aiohttp.ClientSession object')
    if not method.lower() in ('get', 'post', 'put', 'delete', 'patch', 'options', 'head'):
        raise ValueError('method must be one of: get, post, put, delete, patch, options, head. Got: "{}"'.format(method))
    method = method.lower() # type: ignore
    
    # The SSE spec requires making requests with Cache-Control: nocache
    headers = headers or {}
    headers.update(
            {
                'Cache-Control' : 'no-cache',
                'Accept': 'text/event-stream',
            }
        )
    if last_id:
        headers['Last-Event-ID'] = last_id

    # Override default timeout of 5 minutes
    session = session or aiohttp.ClientSession()
    json = json or {}
    async with session:
        try:
            sse_method = getattr(session, method)
            response = await sse_method(url, headers=headers, json=json, **kwargs)
            if response.status not in valid_http_codes:
                await session.close()
                raise ValueError('Invalid HTTP response code: {}'.format(response.status))
                
            response_lines = []
            async for line in response.content:
                line = line.decode('utf-8')
                
                if line in SSESep.__members__.values():  # if line is a separator, means the end of an event
                    if response_lines[0] == ':ok\n':
                        response_lines = []
                        continue

                    current_event = SSEvent.parse(''.join(response_lines))
                    if events and current_event.event in events:
                        try:
                            await _run_func(events[current_event.event], current_event)
                        except Exception as e:
                            _LOGGER.error('Error in event %s: %s', current_event.event, e)
                    yield current_event
                    
                    if exit_events and current_event.event in exit_events:
                        await session.close()
                    response_lines = []
                else:
                    response_lines.append(line)
                
        except TimeoutError as e:
            _LOGGER.error('TimeoutError: %s', e)
        finally:
            if not session.closed:
                await session.close()


__all__ = ['aiosseclient']


'''
Module that fixes SSE prase bugs when facing to single '\n' characters.
Modify from module `aiosseclient`

Example:
```python
# fastapi side

@app.post('/sse')   # support all http methods
async def sse_endpoint(data: dict):
    async def f():
        for i in range(10):
            yield '\n'
            await asyncio.sleep(0.2)
    return EventSourceResponse(f())

##################
# client side

async for event in aiosseclient(url=some_url, method='post', json=some_data):
    print(data, end='', flush=True)   # can get single `'\n'` correctly

```
    
'''

from .datatypes import *

from .client import *


'''
aio sse client
Edit from pypi module `aiosseclient`
'''

import re, logging
from enum import Enum

import sys
if sys.version_info >= (3, 11):
    from typing import List, Optional, Final, Union, Self, Literal
else:
    from typing import List, Optional, Final, Union, Literal
    from typing_extensions import Self

from dataclasses import dataclass

_SSE_LINE_PATTERN: Final[re.Pattern] = re.compile('(?P<name>[^:]*):?( ?(?P<value>.*))?')
_LOGGER = logging.getLogger("aio-sse-chat")


class SSESep(str, Enum):
    '''
    Available separators for SSE.
    SSE only allows these separators.
    '''
    R_N = '\r\n'
    R = '\r'
    N = '\n'
    
    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, DefaultSSEType):
            return self.value == other.value
        return Enum.__eq__(self, other)


class DefaultSSEType(str, Enum):
    '''The default event types for SSE'''
    
    DATA = 'data'
    EVENT = 'event'
    ID = 'id'
    RETRY = 'retry'
    Comment = ':'
    
    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, DefaultSSEType):
            return self.value == other.value
        return Enum.__eq__(self, other)
    
    @classmethod
    def GetSSEType(cls, sse_type_str: str)->Union[Self, str]:
        for member in cls:
            if member.value == sse_type_str:
                return member
        return sse_type_str


@dataclass
class SSEventContent:
    '''Each content means a line in the SSE message. It can be a comment, data, event, id, or retry.'''
    content_type: Union[DefaultSSEType, str]
    value: str


DEFAULT_EVENT_NAME = 'message'
class SSEvent:
    '''The object created as the result of received events'''

    event: str
    id: Optional[str]
    contents: List[SSEventContent]
    retry: Optional[bool]

    def __init__(
        self,
        event: str = DEFAULT_EVENT_NAME,
        id: Optional[str] = None,
        retry: Optional[bool] = None
    ):
        self.event = event
        self.id = id
        self.retry = retry
        self.contents = []

    @classmethod
    def parse(cls, raw:str):
        '''
        Given a possibly-multiline string representing an SSE message, parse it
        and return a Event object.
        '''
        event = cls()
        for line in raw.splitlines():
            m = _SSE_LINE_PATTERN.match(line)
            if m is None:
                # Malformed line.  Discard but warn.
                _LOGGER.warning('Invalid SSE line: %s', line)
                continue

            name = m.group('name')
            if name == '':
                if line[0] == ':':
                    event.contents.append(SSEventContent(DefaultSSEType.Comment, line[1:]))
                continue    # line began with a ':', so is a comment.  Ignore
            else:
                sse_type = DefaultSSEType.GetSSEType(name)
            
            value = m.group('value')
            event.contents.append(SSEventContent(sse_type, value))
            
            if sse_type == DefaultSSEType.EVENT:
                event.event = value
            elif sse_type == DefaultSSEType.ID:
                event.id = value
            elif sse_type == DefaultSSEType.RETRY:
                event.retry = int(value)

        return event

    def __str__(self) -> str:
        return self.data

    @property
    def data(self) -> str:
        data_str = ''
        last_data = None
        for content in self.contents:
            if content.content_type == DefaultSSEType.DATA:
                data_str += (('\n' if last_data is not None else '') + content.value)
                last_data = content.value
        return data_str



__all__ = ['SSEvent', 'SSEventContent', 'DefaultSSEType', 'SSESep']




import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="aiossechat",
  version="0.0.4",
  author="92MING",
  author_email="yashin.sd123@yahoo.com.hk",
  description="A simple sse implementation with aiohttp, which is specifically designed for chatting with LLM.",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/92MING/aiossechat",
  packages=setuptools.find_packages(),
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
)